from pathlib import Path


def update_settings_from_gui(settings, gui):
    _update_main_settings_from_gui(settings, gui)
    _update_flag_settings_from_gui(settings, gui)
    _update_time_settings_from_gui(settings, gui)
    _update_output_settings_from_gui(settings, gui)
    _update_compile_settings_from_gui(settings, gui)
    _update_run_settings_from_gui(settings, gui)
    _update_custom_settings_from_gui(settings, gui)
    _update_rsm_settings_from_gui(settings, gui)


def _update_main_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_MAIN": {
                "work_dir": str(Path(gui.main_dir.text()).resolve()),
                "code_dir": str(
                    Path(gui.main_dir.text()).resolve() / gui.code_dir.text()
                ),
                "case_dir": str(
                    Path(gui.main_dir.text()).resolve() / gui.case_dir.text()
                ),
                "chem_dir": str(
                    Path(gui.main_dir.text()).resolve()
                    / gui.chem_dir.text()
                    / gui.chemname_dir.text()
                ),
                "input_dir": str(Path(gui.input_dir.text()).resolve()),
                "output_dir": str(Path(gui.output_dir.text()).resolve()),
                # Note: always enable trajectory mode
                "station": "traj",
            }
        }
    )

    settings.patch(
        {
            "NML_GUI": {
                "main_casename_dir": str(Path(gui.casename_dir.text())),
                "main_chemall_dir": str(
                    Path(gui.main_dir.text()).resolve() / gui.chem_dir.text()
                ),
                "main_chemname_dir": str(Path(gui.chemname_dir.text())),
            }
        }
    )


def _update_flag_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_FLAG": {
                "flag_emis": 2 if gui.flag_emis.isChecked() else 0,
                "flag_chem": 1 if gui.flag_chem.isChecked() else 0,
                "flag_gasdrydep": 1 if gui.flag_gasdrydep.isChecked() else 0,
                "flag_aero": 1 if gui.flag_aero.isChecked() else 0,
                # Note: Soil emissions are currently always disabled
                "flag_emis_soil": 0,
                "flag_debug": 1 if gui.flag_debug.isChecked() else 0,
                "flag_vapor": 1 if gui.flag_vapor.isChecked() else 0,
                # Note: The output lists are always read from the INITFILE
                "flag_outlist": 0,
                # Note: The format flag is always set to 1
                "flag_format": 1,
                # Note: always enable trajectory mode
                "flag_model_type": 2,
                "flag_mix_chem": 1 if gui.flag_mix_chem.isChecked() else 0,
                "flag_mix_aero": 1 if gui.flag_mix_aero.isChecked() else 0,
                # Note: The aerosols are always simulated in parallel
                "use_parallel_aerosol": True,
            },
            "AER_FLAG": {
                "options": {
                    "nucleation": gui.aer_nucleation.isChecked(),
                    "condensation": gui.aer_condensation.isChecked(),
                    "coagulation": gui.aer_coagulation.isChecked(),
                    "dry_deposition": gui.aer_dry_deposition.isChecked(),
                    # Note: Aerosol wet deposition is not yet implemented
                    # "wet_deposition": gui.aer_wet_deposition.isChecked(),
                    "snow_scavenge": gui.aer_snow_scavenge.isChecked(),
                },
            },
        }
    )


def _update_time_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_TIME": {
                "start_date": [
                    int(i)
                    for i in gui.start_date.dateTime()
                    .toString("yyyy,MM,dd,HH,mm,ss")
                    .split(",")
                ],
                "end_date": [
                    int(i)
                    for i in gui.end_date.dateTime()
                    .toString("yyyy,MM,dd,HH,mm,ss")
                    .split(",")
                ],
                # Note: aero_start_date = start_date + aero_start_offset
                "aero_start_date": [
                    int(i)
                    for i in gui.start_date.dateTime()
                    .addSecs(gui.aero_start_offset.value() * (60 * 60))
                    .toString("yyyy,MM,dd,HH,mm,ss")
                    .split(",")
                ],
                "dt_obs": gui.dt_obs.value(),
                "dt_mete": gui.dt_mete.value(),
                "dt_emis": gui.dt_emis.value(),
                "dt_chem": gui.dt_chem.value(),
                "dt_depo": gui.dt_depo.value(),
                "dt_aero": gui.dt_aero.value(),
                "dt_uhma": gui.dt_uhma.value(),
                # Note: time_zone is always +0 UTC in trajectory mode
                "time_zone": 0.0,
            },
        }
    )


def _update_output_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_OUTPUT": {
                "output_list_spc": ", ".join(
                    s.strip()
                    for s in gui.output_list_spc.toPlainText()
                    .replace("\n", ",")
                    .split(",")
                    if len(s.strip()) > 0
                ),
                "output_list_emi": ", ".join(
                    s.strip()
                    for s in gui.output_list_emi.toPlainText()
                    .replace("\n", ",")
                    .split(",")
                    if len(s.strip()) > 0
                ),
                "output_list_Vd": ", ".join(
                    s.strip()
                    for s in gui.output_list_Vd.toPlainText()
                    .replace("\n", ",")
                    .split(",")
                    if len(s.strip()) > 0
                ),
                "output_list_vap": ", ".join(
                    s.strip()
                    for s in gui.output_list_vap.toPlainText()
                    .replace("\n", ",")
                    .split(",")
                    if len(s.strip()) > 0
                ),
            }
        }
    )

    settings.patch(
        {
            "NML_GUI": {
                "scenario_description": gui.description.toPlainText(),
            }
        }
    )


def _update_compile_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_GUI": {
                "sosaa_exe": str(Path(gui.sosaa_exe.text())),
            }
        }
    )


def _update_run_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_GUI": {
                "launch_cmd": gui.launch_cmd.text(),
            }
        }
    )


def _update_rsm_settings_from_gui(settings, gui):
    settings.patch(
        {
            "NML_RSM": {
                "rsm_path": str(Path(gui.rsm_path.text()).resolve()),
                "train_seed": gui.rsm_train_seed.text(),
                "forest_size": gui.rsm_forest.value(),
                "train_samples": gui.rsm_train_samples.value(),
                "rsm_output": str(Path(gui.rsm_output.text()).resolve()),
                "predict_seed": gui.rsm_predict_seed.text(),
                "predict_samples": gui.rsm_predict_samples.value(),
                "predict_perturbation": gui.rsm_perturbation.toPlainText(),
            }
        }
    )


def _update_custom_settings_from_gui(settings, gui):
    for i in range(1, 31):
        key = getattr(gui, f"customKey_{i}").text().strip()
        value = getattr(gui, f"customVal_{i}").text().strip()

        if key != "" and value != "":
            while True:
                # Integer value
                try:
                    value = int(value)
                    break
                except Exception:
                    pass

                # Real value
                try:
                    value = float(value)
                    break
                except Exception:
                    pass

                # Logical value
                if value.upper() == ".TRUE.":
                    value = True
                    break
                elif value.upper() == ".FALSE.":
                    value = False
                    break

                # Integer array
                try:
                    value = [int(v) for v in value.split(",")]
                    break
                except Exception:
                    pass

                # Real array
                try:
                    value = [float(v) for v in value.split(",")]
                    break
                except Exception:
                    pass

                # String value
                value = str(value)
                break

            settings.patch(
                {
                    "NML_CUSTOM": {
                        key: value,
                    },
                }
            )



from typing import Any, Callable

from PyQt5.QtCore import QObject, QThread, pyqtSignal


def run_in_thread(
    task: Callable[[], Any],
    on_success: Callable[[Any], None],
    on_exception: Callable[[Exception], None],
) -> QThread:
    thread = QThread()
    thread.my_worker = _Worker(task)
    thread.my_worker.moveToThread(thread)

    thread.started.connect(thread.my_worker.run)
    thread.my_worker.on_finished.connect(thread.quit)
    thread.my_worker.on_finished.connect(thread.my_worker.deleteLater)
    thread.finished.connect(thread.deleteLater)

    thread.my_worker.on_success.connect(on_success)
    thread.my_worker.on_exception.connect(on_exception)

    thread.start()

    return thread


class _Worker(QObject):
    on_success = pyqtSignal(object)
    on_exception = pyqtSignal(Exception)
    on_finished = pyqtSignal()

    def __init__(self, task: Callable[[], Any], *args, **kwargs):
        self.task = task

        super().__init__(*args, **kwargs)

    def run(self):
        try:
            result = self.task()
            self.on_success.emit(result)
        except Exception as err:
            self.on_exception.emit(err)
        finally:
            self.on_finished.emit()



from ..netcdf import TrajectoryDatasets
from ..utils import df_to_numpy
from .ccn import get_ccn_concentration


def get_labels_for_dataset(ds: TrajectoryDatasets):
    import numpy as np
    import pandas as pd

    ccn_concentration = get_ccn_concentration(ds)

    ccn_concentration_np = df_to_numpy(ccn_concentration)

    labels_np = np.concatenate(
        [
            ccn_concentration.index.get_level_values(0)
            .to_numpy()
            .reshape(
                (
                    ccn_concentration.index.levels[0].size,
                    ccn_concentration.index.levels[1].size,
                    1,
                )
            ),
            ccn_concentration.index.get_level_values(1)
            .to_numpy()
            .reshape(
                (
                    ccn_concentration.index.levels[0].size,
                    ccn_concentration.index.levels[1].size,
                    1,
                )
            ),
            ccn_concentration_np.reshape(
                (
                    ccn_concentration_np.shape[0],
                    ccn_concentration_np.shape[1],
                    1,
                )
            ),
        ],
        axis=2,
    )

    # Trim off the first two days, for which the time features are ill-defined
    labels_np_trimmed = labels_np[96:, :, :]

    label_names = ["time", "level", "ccn"]

    labels = pd.DataFrame(
        labels_np_trimmed.reshape(
            labels_np_trimmed.shape[0] * labels_np_trimmed.shape[1],
            labels_np_trimmed.shape[2],
        ),
        columns=label_names,
    ).set_index(["time", "level"])

    return labels



import datetime

from PyQt5 import QtWidgets


def init_plot(gui):
    if getattr(gui, "rsm_diffplot_canvas", None) is not None:
        return

    import matplotlib as mpl
    from matplotlib import pyplot as plt
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
    from matplotlib.backends.backend_qt5agg import (
        NavigationToolbar2QT as NavigationToolbar,
    )

    gui.rsm_diffplot_fig, gui.rsm_diffplot_ax = plt.subplots(
        1, 1, figsize=(6, 4), dpi=100
    )
    gui.rsm_diffplot_canvas = FigureCanvasQTAgg(gui.rsm_diffplot_fig)

    gui.rsm_diffplot_cb = gui.rsm_diffplot_fig.colorbar(
        mpl.cm.ScalarMappable(
            norm=None,
            cmap="viridis",
        ),
        ax=gui.rsm_diffplot_ax,
        orientation="horizontal",
        extend="min",
    )

    diffplot_layout = QtWidgets.QVBoxLayout()
    diffplot_layout.addWidget(NavigationToolbar(gui.rsm_diffplot_canvas, gui))
    diffplot_layout.addWidget(gui.rsm_diffplot_canvas)

    gui.rsm_diffplot_tab.setLayout(diffplot_layout)


def update_plot(gui):
    import numpy as np

    gui.rsm_diffplot_ax.cla()

    gui.rsm_diffplot_ax.set_title(
        "Change in Cloud-Condensation-Nuclei (CCN) Concentration"
    )

    gui.rsm_diffplot_ax.set_xlabel("Baseline CCN concentration [m$^{-3}$]")
    gui.rsm_diffplot_ax.set_ylabel(
        "Perturbed Change in CCN concentration [m$^{-3}$]"
    )

    if gui.rsm_prediction is None:
        gui.rsm_diffplot_ax.set_xscale("linear")
        gui.rsm_diffplot_ax.set_yscale("linear")

        gui.rsm_diffplot_ax.set_xticks([])
        gui.rsm_diffplot_ax.set_yticks([])

        gui.rsm_diffplot_cb.ax.set_xticks([])

        gui.rsm_diffplot_ax.text(
            0.5,
            0.5,
            "Missing SOSAA RSM Predictions",
            color="red",
            size=20,
            ha="center",
            va="center",
            transform=gui.rsm_diffplot_ax.transAxes,
        )
    else:
        time = gui.rsm_prediction.index.get_level_values(0)
        log10_ccn_baseline = (
            gui.rsm_prediction["log10_ccn_baseline"].to_numpy().flatten()
        )
        log10_ccn_perturbed_pred = (
            gui.rsm_prediction["log10_ccn_perturbed_pred"].to_numpy().flatten()
        )
        log10_ccn_perturbed_stdv = (
            gui.rsm_prediction["log10_ccn_perturbed_stdv"].to_numpy().flatten()
        )
        log10_ccn_perturbed_conf = (
            gui.rsm_prediction["log10_ccn_perturbed_conf"].to_numpy().flatten()
        )

        gui.rsm_diffplot_ax.set_title(
            "Change in Cloud-Condensation-Nuclei (CCN) Concentration"
            + f"\nConfidence: {np.mean(log10_ccn_perturbed_conf):.02}"
            + "\nWARNING: The SOSAA RSM is currently known to struggle"
            + " with small perturbations."
        )

        for _ in range(gui.rsm_train_samples.value()):
            log10_ccn_perturbed = np.random.normal(
                loc=log10_ccn_perturbed_pred,
                scale=log10_ccn_perturbed_stdv,
            )

            I_conf = (
                np.random.random(size=log10_ccn_perturbed_conf.shape)
                <= log10_ccn_perturbed_conf
            )

            gui.rsm_diffplot_ax.scatter(
                np.power(10.0, log10_ccn_baseline[I_conf]),
                np.power(10.0, log10_ccn_perturbed[I_conf])
                - np.power(10.0, log10_ccn_baseline[I_conf]),
                c=time[I_conf],
                cmap="viridis",
                s=5,
            )

        xlim = gui.rsm_diffplot_ax.get_xlim()
        gui.rsm_diffplot_ax.plot(xlim, [0, 0], c="black", lw=1)
        gui.rsm_diffplot_ax.set_xlim(xlim)

        gui.rsm_diffplot_cb.ax.set_xticks(
            [
                (
                    (
                        h
                        - int(
                            (
                                gui.rsm_prediction.index.levels[0][0]
                                // (60 * 60)
                            )
                        )
                    )
                    / (
                        -int(
                            (
                                gui.rsm_prediction.index.levels[0][0]
                                // (60 * 60)
                            )
                        )
                    )
                )
                for h in range(
                    int((gui.rsm_prediction.index.levels[0][0] // (60 * 60))),
                    0,
                    24,
                )
            ]
        )
        gui.rsm_diffplot_cb.ax.set_xticklabels(
            [
                (gui.rsm_dt + datetime.timedelta(hours=h)).strftime("%d.%m")
                for h in range(
                    int((gui.rsm_prediction.index.levels[0][0] // (60 * 60))),
                    0,
                    24,
                )
            ]
        )

    gui.rsm_diffplot_fig.tight_layout()
    gui.rsm_diffplot_canvas.draw()



def init_modules_gui(self):
    def changeFlagEmis(flag_emis):
        self.dt_emis.setEnabled(flag_emis)
        self.label_dt_emis.setEnabled(flag_emis)
        self.group_output_list_emi.setEnabled(flag_emis)

    self.flag_emis.stateChanged.connect(changeFlagEmis)

    def changeFlagChem(flag_chem):
        self.dt_chem.setEnabled(flag_chem)
        self.label_dt_chem.setEnabled(flag_chem)
        self.label_chem_dir.setEnabled(flag_chem)
        self.chem_dir.setEnabled(flag_chem)
        self.browse_chem.setEnabled(flag_chem)

    self.flag_chem.clicked.connect(changeFlagChem)

    def changeFlagGasDryDep(flag_gasdrydep):
        self.dt_depo.setEnabled(flag_gasdrydep)
        self.label_dt_depo.setEnabled(flag_gasdrydep)
        self.group_output_list_Vd.setEnabled(flag_gasdrydep)

    self.flag_gasdrydep.stateChanged.connect(changeFlagGasDryDep)

    def changeFlagAero(flag_aero):
        self.dt_aero.setEnabled(flag_aero)
        self.label_dt_aero.setEnabled(flag_aero)
        self.dt_uhma.setEnabled(flag_aero)
        self.label_dt_uhma.setEnabled(flag_aero)
        self.aero_start_label.setEnabled(flag_aero)
        self.aero_start_offset.setEnabled(flag_aero)
        self.aero_start_date.setEnabled(flag_aero)

    self.flag_aero.clicked.connect(changeFlagAero)
